from typing import Any

from prefect.engine.result import SafeResult
from prefect.engine.result_handlers import ResultHandler


class PurgedResultType(SafeResult):
    """
    A `SafeResult` subclass representing a Result that has been purged.  A `PurgedResult` object
    returns itself for its `value` and its `safe_value`.
    """

    def __init__(self) -> None:
        super().__init__(value=None, result_handler=ResultHandler())

    def __eq__(self, other: Any) -> bool:
        if type(self) == type(other):
            return True
        else:
            return False

    def __repr__(self) -> str:
        return "<Purged result>"

    def __str__(self) -> str:
        return "PurgedResult"

    def to_result(self, result_handler: ResultHandler = None) -> "ResultInterface":
        """
        Performs no computation and returns self.

        Args:
            - result_handler (optional): a passthrough for interface compatibility
        """
        return self


PurgedResult = PurgedResultType()



import os
import typing

from prefect.engine.state import State, Success
from prefect.engine.result import Result
from prefect.core.edge import Edge

from prefect_ds.task_runner import DSTaskRunner


def checkpoint_handler(task_runner: DSTaskRunner, old_state: State, new_state: State) -> State:
    """
    A handler designed to implement result caching by filename. If the result handler's ``read``
    method can be successfully run, this handler loads the result of that method as the task result
    and sets the task state to ``Success``. Similarly, on successful
    completion of the task, if the task was actually run and not loaded from cache, this handler
    will apply the result handler's ``write`` method to the task.

    Parameters
    ----------
    task_runner : instance of DSTaskRunner
        The task runner associated with the flow the handler is used in.
    old_state : instance of prefect.engine.state.State
        The current state of the task.
    new_state : instance of prefect.engine.state.State
        The expected new state of the task.

    Returns
    -------
    new_state : instance of prefect.engine.state.State
        The actual new state of the task.
    """
    if "PREFECT__FLOWS__CHECKPOINTING" in os.environ and os.environ["PREFECT__FLOWS__CHECKPOINTING"] == "true":
        raise AttributeError("Cannot use standard prefect checkpointing with this handler")

    if task_runner.result_handler is not None and old_state.is_pending() and new_state.is_running():
        if not hasattr(task_runner, "upstream_states"):
            raise TypeError(
                "upstream_states not found in task runner. Make sure to use "
                "prefect_ds.task_runner.DSTaskRunner."
            )
        input_mapping = _create_input_mapping(task_runner.upstream_states)
        try:
            data = task_runner.task.result_handler.read(input_mapping=input_mapping)
        except FileNotFoundError:
            return new_state
        except TypeError: # unexpected argument input_mapping
            raise TypeError(
                "Result handler could not accept input_mapping argument. "
                "Please ensure that you are using a handler from prefect_ds."
            )
        result = Result(value=data, result_handler=task_runner.task.result_handler)
        state = Success(result=result, message="Task loaded from disk.")
        return state

    if task_runner.result_handler is not None and old_state.is_running() and new_state.is_successful():
        input_mapping = _create_input_mapping(task_runner.upstream_states)
        task_runner.task.result_handler.write(new_state.result, input_mapping=input_mapping)

    return new_state


def _create_input_mapping(upstream_states: typing.Dict[Edge, State]) -> typing.Dict[str, typing.Any]:
    mapping = {}
    for edge, state in upstream_states.items():
        input_variable_name = edge.key
        input_task_result = state.result
        mapping[input_variable_name] = input_task_result
    return mapping


from prefect.core import Edge
from prefect.engine.state import State
from prefect.engine.task_runner import TaskRunner
from typing import Dict, Any


class DSTaskRunner(TaskRunner):
    def run(
        self,
        state: State = None,
        upstream_states: Dict[Edge, State] = None,
        context: Dict[str, Any] = None,
        executor: "prefect.engine.executors.Executor" = None,
    ) -> State:
        """
        See the documentation for ``prefect.engine.task_runner.TaskRunner.run()``.
        """
        self.upstream_states = upstream_states
        return super().run(state=state, upstream_states=upstream_states, context=context, executor=executor)



import pandas as pd
import pathlib
import typing

from prefect.engine.result_handlers.result_handler import ResultHandler


def _generate_pandas_io_methods() -> typing.Tuple[typing.Dict[str, typing.Callable], typing.Dict[str, str]]:
    all_read_ops = [method for method in dir(pd) if method.lower().startswith("read_")]
    all_write_ops = [method for method in dir(pd.DataFrame()) if method.lower().startswith("to_")]
    read_suffixes = [method.lower().split("_")[1] for method in all_read_ops]
    write_suffixes = [method.lower().split("_")[1] for method in all_write_ops]

    read_io_ops = {
        read_method.split("_")[1].lower(): getattr(pd, read_method)
        for read_method in all_read_ops
        if read_method.split("_")[1].lower() in write_suffixes
    }
    write_io_ops = {
        write_method.split("_")[1].lower(): write_method
        for write_method in all_write_ops
        if write_method.split("_")[1].lower() in read_suffixes
    }

    return read_io_ops, write_io_ops


class PandasResultHandler(ResultHandler):
    """
    Hook for storing and retrieving task results in Pandas DataFrames.
    Task results are written/read via the standard Pandas
    ``to_[FILETYPE]``/``read_[FILETYPE]`` methods, and additionally the filename
    can be fully specified at task instantiation.

    Parameters
    ----------
    path : str or pathlib.Path
        Filepath to be read from or written to, including file name and extension.
    file_type : str
        The type of file to write to, e.g. "csv" or "parquet". Must match the name
        of the appropriate ``to_[FILETYPE]``/``read_[FILETYPE]`` method.
    read_kwargs : dict or None
        If present, passed as **kwargs to the ``read_[FILETYPE]`` method.
    write_kwargs : dict or None
        If present, passed as **kwargs tot he ``to_[FILETYPE]`` method.

    .. note::
        Because the filepath is fully specified, when using this handler in a ``map``
        can lead to the same file being read to or written from by every iteration of
        the map. To deal with this case this handler allows for Python's ``str.format``
        to be used with the names of arguments to the task instance. For instance, if
        your task has an argument named ``sample_name`` that you intend to map over,
        supplying a ``file_type`` like ``"output_{sample_name}.csv"`` will fill in the
        values of that argument for each iteration of the map.
    """
    _READ_OPS_MAPPING, _WRITE_OPS_MAPPING = _generate_pandas_io_methods()
    assert set(_READ_OPS_MAPPING.keys()) == set(_WRITE_OPS_MAPPING.keys())

    def __init__(
            self,
            path: typing.Union[str, pathlib.Path],
            file_type: str,
            read_kwargs: dict = None,
            write_kwargs: dict = None
    ):
        self.path = pathlib.Path(path)
        self.file_type = file_type

        if self.file_type.lower() not in self._READ_OPS_MAPPING:
            raise ValueError(
                f"{self.file_type} not available. "
                f"Known file extensions are {list(self._READ_OPS_MAPPING.keys())}"
            )
        self.read_kwargs = read_kwargs if read_kwargs is not None else {}
        self.write_kwargs = write_kwargs if write_kwargs is not None else {}
        super().__init__()

    def read(self, *, input_mapping=None) -> pd.DataFrame:
        """
        Read a result from the specified ``path`` using the appropriate ``read_[FILETYPE]`` method.

        Parameters
        ----------
        input_mapping : dict
            If present, passed to ``path.format()`` to set the final filename. This is necessary
            for mapped tasks, to prevent the same file from being read from for each map sub-step.
        """

        input_mapping = {} if input_mapping is None else input_mapping
        path_string = str(self.path).format(**input_mapping)
        self.logger.debug("Starting to read result from {}...".format(path_string))
        data = self._READ_OPS_MAPPING[self.file_type.lower()](
            path_string,
            **self.read_kwargs
        )
        self.logger.debug("Finished reading result from {}...".format(path_string))
        return data

    def write(self, result: pd.DataFrame, input_mapping=None):
        input_mapping = {} if input_mapping is None else input_mapping
        path_string = str(self.path).format(**input_mapping)
        self.logger.debug("Starting to write result to {}...".format(path_string))
        write_function = getattr(result, self._WRITE_OPS_MAPPING[self.file_type.lower()])
        write_function(path_string, **self.write_kwargs)
        self.logger.debug("Finished writing result to {}...".format(path_string))



from prefect.engine.flow_runner import FlowRunner
from prefect.engine.state import State
from prefect.core.edge import Edge
from prefect.core.task import Task
from typing import Any, Callable, Dict, Iterable, Set

from prefect_ds.result import PurgedResult


class DSFlowRunner(FlowRunner):

    def get_flow_run_state(
        self,
        state: State,
        task_states: Dict[Task, State],
        task_contexts: Dict[Task, Dict[str, Any]],
        return_tasks: Set[Task],
        task_runner_state_handlers: Iterable[Callable],
        executor: "prefect.engine.executors.base.Executor",
    ) -> State:
        """
        See the documentation for ``prefect.engine.flow_runner.FlowRunner.get_flow_run_state()``.
        """
        self.task_states = task_states
        return super().get_flow_run_state(
            state=state,
            task_states=task_states,
            task_contexts=task_contexts,
            return_tasks=return_tasks,
            task_runner_state_handlers=task_runner_state_handlers,
            executor=executor
        )

    def run_task(
        self,
        task: Task,
        state: State,
        upstream_states: Dict[Edge, State],
        context: Dict[str, Any],
        task_runner_state_handlers: Iterable[Callable],
        executor: "prefect.engine.executors.Executor",
    ) -> State:
        """
        Same as ``prefect.engine.flow_runner.FlowRunner.run_task()``, but checks
        to see if upstream tasks can be purged after the current task is run.
        """
        task_output = super().run_task(
            task=task,
            state=state,
            upstream_states=upstream_states,
            context=context,
            task_runner_state_handlers=task_runner_state_handlers,
            executor=executor
        )
        self._purge_unnecessary_tasks(task, upstream_states)
        return task_output

    def _purge_unnecessary_tasks(self, task, upstream_state_edges):
        for state_edge in upstream_state_edges:
            upstream_task = state_edge.upstream_task
            #breakpoint()
            try:
                edges_from_upstream_task = self.flow.edges_from(upstream_task)
            except ValueError:
                # When a non-task (e.g. an int) is used as a task input,
                # it doesn't show up in the task list. In these cases, we
                # don't need to worry about purging anything.
                continue

            is_safely_purgeable = True
            for edge in edges_from_upstream_task:
                if edge.downstream_task == task:
                    # Downstream task is the current task
                    continue
                if (
                        edge.downstream_task in self.task_states and
                        self.task_states[edge.downstream_task].is_successful() == True
                ):
                    # Downstream edge has been successfully completed
                    continue
                # If we're still in this iteration of the loop, it means that this
                # downstream task hasn't successfully completed yet
                is_safely_purgeable = False
                break
            if is_safely_purgeable:
                if self.task_states[upstream_task].is_mapped():
                    for mapped_state in self.task_states[upstream_task].map_states:
                        mapped_state._result = PurgedResult
                self.task_states[upstream_task]._result = PurgedResult
